"""
Helpers for scanning SEGY files by shot location.
"""

from typing import Any, Dict, Iterable, List, Optional, Tuple
from concurrent.futures import ThreadPoolExecutor, as_completed
import os
import threading

import fnmatch
from dataclasses import dataclass, field
import struct
import numpy as np
import cloudpickle

from .read import read_fileheader, read_traceheader, read_traces
from .utils import get_header
from .types import (
    SeisBlock,
    FileHeader,
    BinaryTraceHeader,
    TH_BYTE2SAMPLE,
)


@dataclass
class ShotRecord:
    """Information about a single shot location within a SEGY file."""

    path: str
    coordinates: Tuple[float, float, float]
    fileheader: FileHeader
    rec_depth_key: str = "GroupWaterDepth"
    segments: List[Tuple[int, int]] = field(default_factory=list)
    summary: dict = field(default_factory=dict)
    ns: int = 0
    dt: int = 0
    fs: Any = field(default=None, repr=False)
    _data: Optional[np.ndarray] = field(default=None, init=False, repr=False)
    _headers: Optional[List[BinaryTraceHeader]] = field(
        default=None, init=False, repr=False
    )
    _rec_coords: Optional[np.ndarray] = field(
        default=None, init=False, repr=False
    )

    def __str__(self) -> str:
        lines = ["ShotRecord:"]
        lines.append(f"    path: {self.path}")
        lines.append(
            "    source: ("
            f"{self.coordinates[0]}, {self.coordinates[1]}, "
            f"{self.coordinates[2]}"
            ")"
        )
        lines.append(f"    traces: {sum(c for _, c in self.segments)}")
        lines.append(f"    ns: {self.ns}, dt: {self.dt}")
        if self.summary:
            lines.append("    summary:")
            for k, (mn, mx) in self.summary.items():
                lines.append(f"        {k:30s}: {mn}..{mx}")
        return "\n".join(lines)

    __repr__ = __str__

    def read_data(self, keys: Optional[Iterable[str]] = None) -> SeisBlock:
        """Load all traces for this shot."""
        data_parts = []
        for offset, count in self.segments:
            opener = self.fs.open if self.fs is not None else open
            with opener(self.path, "rb") as f:
                f.seek(offset)
                h, d = read_traces(
                    f,
                    self.fileheader.bfh.ns,
                    count,
                    self.fileheader.bfh.DataSampleFormat,
                    keys,
                )
                data_parts.append(d)
        return np.concatenate(data_parts, axis=0) if data_parts else []

    def read_headers(
        self, keys: Optional[Iterable[str]] = None
    ) -> List[BinaryTraceHeader]:
        """Read only the headers for this shot."""
        headers: List[BinaryTraceHeader] = []
        ns = self.fileheader.bfh.ns
        for offset, count in self.segments:
            opener = self.fs.open if self.fs is not None else open
            with opener(self.path, "rb") as f:
                f.seek(offset)
                for _ in range(count):
                    th = read_traceheader(f, keys)
                    headers.append(th)
                    f.seek(ns * 4, os.SEEK_CUR)
        return headers

    @property
    def data(self) -> SeisBlock:
        if self._data is None:
            self._data = self.read_data()
        return self._data

    @property
    def rec_coordinates(self) -> np.ndarray:
        """Array of receiver coordinates for this shot."""
        if self._rec_coords is None:
            hdrs = self.read_headers(
                keys=[
                    "GroupX",
                    "GroupY",
                    self.rec_depth_key,
                    "RecSourceScalar",
                    "ElevationScalar",
                ]
            )
            gx = get_header(hdrs, "GroupX")
            gy = get_header(hdrs, "GroupY")
            dz = get_header(hdrs, self.rec_depth_key)
            self._rec_coords = np.column_stack((gx, gy, dz)).astype(np.float32)
        return self._rec_coords


def _parse_header(buf: bytes, keys: Iterable[str]) -> BinaryTraceHeader:
    """
    Return a :class:`BinaryTraceHeader` parsed from ``buf``.

    Parameters
    ----------
    buf : bytes
        240-byte buffer containing the raw trace header.
    keys : Iterable[str]
        Header fields to decode from ``buf``.

    Returns
    -------
    BinaryTraceHeader
        Trace header populated with the requested fields.
    """
    th = BinaryTraceHeader()
    for k in keys:
        offset, size = TH_BYTE2SAMPLE[k]
        fmt = ">i" if size == 4 else ">h"
        val = struct.unpack_from(fmt, buf, offset)[0]
        setattr(th, k, val)
    th.keys_loaded = list(keys)
    return th


def _update_summary(
    summary: Dict[str, Tuple[float, float]],
    th: BinaryTraceHeader,
    keys: Iterable[str],
) -> None:
    """
    Update ``summary`` with values from ``th``.

    Parameters
    ----------
    summary : dict
        Mapping of header name to ``(min, max)`` tuple.
    th : BinaryTraceHeader
        Header providing new values.
    keys : Iterable[str]
        Header fields to include in the summary.
    """
    for k in keys:
        v = get_header([th], k)[0]
        if k in summary:
            mn, mx = summary[k]
            if v < mn:
                mn = v
            if v > mx:
                mx = v
            summary[k] = (mn, mx)
        else:
            summary[k] = (v, v)


def _iter_trace_headers(
    f,
    start: int,
    count: int,
    ns: int,
    keys: Iterable[str],
    chunk: int = 1024,
) -> Iterable[Tuple[int, BinaryTraceHeader]]:
    """
    Yield offsets and headers from ``f`` starting at ``start``.

    Parameters
    ----------
    f : file-like object
        Opened file positioned at ``start``.
    start : int
        Byte offset of the first trace.
    count : int
        Number of traces to read.
    ns : int
        Samples per trace.
    keys : Iterable[str]
        Header fields to decode.
    chunk : int, optional
        Number of traces to read per block.

    Yields
    ------
    tuple
        ``(offset, header)`` for each trace encountered.
    """
    trace_size = 240 + ns * 4
    pos = start
    remaining = count
    while remaining > 0:
        n = min(chunk, remaining)
        buf = f.read(trace_size * n)
        for i in range(n):
            base = i * trace_size
            hdr = _parse_header(buf[base:base + 240], keys)
            yield pos + base, hdr
        pos += n * trace_size
        remaining -= n


class SegyScan:
    """
    Representation of SEGY data grouped by shot.

    Parameters
    ----------
    fh : FileHeader
        File header shared by all scanned files.
    records : list of ShotRecord
        Collection of shot metadata describing trace segments.
    """

    def __init__(self, fh: FileHeader, records: List[ShotRecord], fs=None) -> None:
        """Create a new :class:`SegyScan` instance.

        Parameters
        ----------
        fh : FileHeader
            File header common to all files being scanned.
        records : list of ShotRecord
            Shot metadata describing trace segments.
        fs : filesystem-like object, optional
            Filesystem providing ``open`` for reading data lazily.
        """
        self.fileheader = fh
        self.records = records
        self.fs = fs
        self._data: Optional[List[SeisBlock]] = None

    def __len__(self) -> int:
        """Return the number of distinct shots."""
        return len(self.records)

    @property
    def paths(self) -> List[str]:
        """List of file paths corresponding to each shot."""
        return [r.path for r in self.records]

    @property
    def shots(self) -> List[Tuple[int, int, int]]:
        """Source coordinates for each shot including depth."""
        return [r.coordinates for r in self.records]

    @property
    def offsets(self) -> List[int]:
        """First trace byte offset for every shot."""
        return [r.segments[0][0] for r in self.records]

    @property
    def counts(self) -> List[int]:
        """Total number of traces for each shot."""
        return [sum(c for _, c in r.segments) for r in self.records]

    def __getitem__(self, idx: int) -> ShotRecord:
        """Return the ``idx``-th :class:`ShotRecord`."""
        return self.records[idx]

    @property
    def data(self) -> List[SeisBlock]:
        """Load data for all shots on first access."""
        if self._data is None:
            self._data = [self.read_data(i) for i in range(len(self.records))]
        return self._data

    def summary(self, idx: int) -> dict:
        """Header summaries for the ``idx``-th shot."""
        return self.records[idx].summary

    def read_data(
        self, idx: int, keys: Optional[Iterable[str]] = None
    ) -> SeisBlock:
        """
        Load all traces for a single shot.

        Parameters
        ----------
        idx : int
            Index of the shot to read.
        keys : Iterable[str], optional
            Additional header fields to load with each trace.

        Returns
        -------
        SeisBlock
            In-memory representation of the selected shot.
        """
        rec = self.records[idx]
        headers: List[BinaryTraceHeader] = []
        data_parts = []
        for offset, count in rec.segments:
            opener = rec.fs.open if rec.fs is not None else (
                self.fs.open if getattr(self, "fs", None) is not None else open
            )
            with opener(rec.path, "rb") as f:
                f.seek(offset)
                h, d = read_traces(
                    f,
                    self.fileheader.bfh.ns,
                    count,
                    self.fileheader.bfh.DataSampleFormat,
                    keys,
                )
                headers.extend(h)
                data_parts.append(d)
        # Concatenate data parts if necessary
        if data_parts:
            data = np.concatenate(data_parts, axis=0)
        else:
            data = []
        return SeisBlock(self.fileheader, headers, data)

    def read_headers(
        self, idx: int, keys: Optional[Iterable[str]] = None
    ) -> List[BinaryTraceHeader]:
        """
        Read only the headers for a single shot.

        Parameters
        ----------
        idx : int
            Shot index to read.
        keys : Iterable[str], optional
            Header fields to populate; by default all are read.

        Returns
        -------
        list of BinaryTraceHeader
            Parsed headers for the requested shot.
        """
        rec = self.records[idx]
        headers: List[BinaryTraceHeader] = []
        ns = self.fileheader.bfh.ns
        for offset, count in rec.segments:
            opener = rec.fs.open if rec.fs is not None else (
                self.fs.open if getattr(self, "fs", None) is not None else open
            )
            with opener(rec.path, "rb") as f:
                f.seek(offset)
                for _ in range(count):
                    th = read_traceheader(f, keys)
                    headers.append(th)
                    f.seek(ns * 4, os.SEEK_CUR)
        return headers

    def __str__(self) -> str:
        lines = ["SegyScan:"]
        lines.append(f"    shots: {len(self.records)}")
        lines.append(f"    ns: {self.fileheader.bfh.ns}")
        lines.append(f"    dt: {self.fileheader.bfh.dt}")
        return "\n".join(lines)

    __repr__ = __str__


def _scan_file(
    path: str,
    keys: Optional[Iterable[str]] = None,
    chunk: int = 1024,
    depth_key: str = "SourceDepth",
    rec_depth_key: str = "GroupWaterDepth",
    fs=None,
) -> SegyScan:
    """
    Scan ``path`` for shot locations.

    Parameters
    ----------
    path : str
        SEGY file to scan.
    keys : Iterable[str], optional
        Additional header fields to summarise.
    chunk : int, optional
        Number of traces to read at once.
    depth_key : str, optional
        Trace header field giving the source depth.
    rec_depth_key : str, optional
        Header field giving the receiver depth.

    fs : filesystem-like object, optional
        Filesystem providing ``open`` if reading from non-local storage.

    Returns
    -------
    SegyScan
        Object describing all shots found in ``path``.
    """
    thread = threading.current_thread().name
    print(f"{thread} scanning file {path}")
    trace_keys = [
        "SourceX",
        "SourceY",
        depth_key,
        "GroupX",
        "GroupY",
        rec_depth_key,
        "RecSourceScalar",
        "ElevationScalar",
    ]
    if keys is not None:
        for k in keys:
            if k not in trace_keys:
                trace_keys.append(k)

    opener = fs.open if fs is not None else open

    with opener(path, "rb") as f:
        fh = read_fileheader(f)
        print(f"Header for {path}: ns={fh.bfh.ns} dt={fh.bfh.dt}")
        ns = fh.bfh.ns
        f.seek(0, os.SEEK_END)
        total = (f.tell() - 3600) // (240 + ns * 4)
        f.seek(3600)

        records: Dict[Tuple[int, int, int], ShotRecord] = {}

        previous: Optional[Tuple[int, int, int]] = None
        seg_start = 0
        seg_count = 0

        for offset, th in _iter_trace_headers(
            f,
            3600,
            total,
            ns,
            trace_keys,
            chunk,
        ):
            src = (
                np.float32(get_header([th], "SourceX")[0]),
                np.float32(get_header([th], "SourceY")[0]),
                np.float32(get_header([th], depth_key)[0]),
            )

            rec = records.get(src)
            if rec is None:
                rec = ShotRecord(
                    path,
                    src,
                    fh,
                    rec_depth_key,
                    [],
                    {},
                    ns,
                    fh.bfh.dt,
                    fs,
                )
                records[src] = rec
            _update_summary(rec.summary, th, keys or [])

            # New segment begins when the source position changes
            if previous is None:
                previous = src
                seg_start = offset
                seg_count = 1
            elif src == previous:
                seg_count += 1
            else:
                records[previous].segments.append((seg_start, seg_count))
                previous = src
                seg_start = offset
                seg_count = 1

        if previous is not None:
            # Append the final segment for the last shot
            records[previous].segments.append((seg_start, seg_count))

    record_list = sorted(records.values(), key=lambda r: r.coordinates)
    print(f"{thread} found {len(record_list)} shots in {path}")
    return SegyScan(fh, record_list, fs=fs)


def segy_scan(
    path: str,
    file_key: Optional[str] = None,
    keys: Optional[Iterable[str]] = None,
    chunk: int = 1024,
    depth_key: str = "SourceDepth",
    rec_depth_key: str = "GroupWaterDepth",
    threads: Optional[int] = None,
    fs=None,
) -> SegyScan:
    """
    Scan one or more SEGY files and merge the results.

    Parameters
    ----------
    path : str
        Directory containing SEGY files or a single file path.
    file_key : str, optional
        Glob pattern selecting files within ``path``. When omitted and
        ``path`` points to a file, only that file is scanned.
    keys : Iterable[str], optional
        Additional header fields to summarise while scanning.
    chunk : int, optional
        Number of traces to read per block.
    depth_key : str, optional
        Header name containing the source depth.
    rec_depth_key : str, optional
        Header field containing the receiver depth.
    fs : filesystem-like object, optional
        Filesystem providing ``open`` and ``glob`` if scanning non-local paths.

    Returns
    -------
    SegyScan
        Combined scan object describing all detected shots.
    """

    if threads is None:
        threads = os.cpu_count() or 1

    if file_key is None and (
        (fs is None and os.path.isfile(path)) or (fs and fs.isfile(path))
    ):
        files = [path]
        if fs is None:
            directory = os.path.dirname(path) or "."
        else:
            directory = getattr(fs, "_parent", os.path.dirname)(path)
    else:
        directory = path
        pattern = file_key or "*"
        if fs is None:
            files = [
                os.path.join(directory, fname)
                for fname in os.listdir(directory)
                if fnmatch.fnmatch(fname, pattern)
            ]
        else:
            files = fs.glob(f"{directory.rstrip('/')}/{pattern}")
    files.sort()

    print(
        f"Scanning {len(files)} files in {directory} with {threads} threads"
    )
    records: List[ShotRecord] = []
    with ThreadPoolExecutor(max_workers=threads) as pool:
        futures = {
            pool.submit(
                _scan_file, f, keys, chunk, depth_key, rec_depth_key, fs
            ): f
            for f in files
        }
        for fut in as_completed(futures):
            scan = fut.result()
            fh = scan.fileheader
            records.extend(scan.records)

    if not records:
        raise FileNotFoundError("No matching SEGY files found")

    records.sort(key=lambda r: r.coordinates)

    print(f"Combined scan has {len(records)} shots")
    return SegyScan(fh, records, fs=fs)


def save_scan(path: str, scan: SegyScan, fs=None) -> None:
    """Serialize ``scan`` to ``path``.

    Parameters
    ----------
    path : str
        Destination file path. When ``fs`` is provided the path is interpreted
        relative to that filesystem.
    scan : SegyScan
        Object to serialize.
    fs : filesystem-like object, optional
        Filesystem providing ``open`` when writing to non-local storage.
    """
    print(f"Saving SegyScan to {path}")
    opener = fs.open if fs is not None else open
    with opener(path, "wb") as f:
        cloudpickle.dump(scan, f, protocol=cloudpickle.DEFAULT_PROTOCOL)
    print(f"Finished saving {path}")


def load_scan(path: str, fs=None) -> SegyScan:
    """Load a :class:`SegyScan` previously saved with :func:`save_scan`.

    Parameters
    ----------
    path : str
        File system path of the saved object. When ``fs`` is provided the path
        is interpreted relative to that filesystem.
    fs : filesystem-like object, optional
        Filesystem providing ``open`` when reading from non-local storage.

    Returns
    -------
    SegyScan
        Deserialized scan object.
    """
    print(f"Loading SegyScan from {path}")
    opener = fs.open if fs is not None else open
    with opener(path, "rb") as f:
        scan = cloudpickle.load(f)

    # When loading from external storage the filesystem won't be part of the
    # serialized object. Attach it so lazy reads work correctly.
    if fs is not None:
        scan.fs = fs
        for rec in scan.records:
            rec.fs = fs

    print(f"Loaded SegyScan with {len(scan.records)} shots")
    return scan



"""
Reading utilities for the minimal Python SEGY implementation.
"""

from .types import (
    BinaryFileHeader,
    BinaryTraceHeader,
    FileHeader,
    SeisBlock,
    FH_BYTE2SAMPLE,
    TH_BYTE2SAMPLE,
)
from typing import BinaryIO, Iterable, List, Optional, Tuple
import numpy as np
from concurrent.futures import ThreadPoolExecutor

from .ibm import ibm_to_ieee_array
import struct

# Number of traces to read at a time when loading an entire file
TRACE_CHUNKSIZE = 512


def read_fileheader(
    f: BinaryIO, keys: Optional[Iterable[str]] = None, bigendian: bool = True
) -> FileHeader:
    """
    Read and parse the binary file header.

    Parameters
    ----------
    f : BinaryIO
        Open binary file handle.
    keys : Iterable[str], optional
        Header fields to read; by default all are loaded.
    bigendian : bool, optional
        ``True`` when the file is big-endian, ``False`` otherwise.

    Returns
    -------
    FileHeader
        Object containing the textual and binary headers.
    """
    if keys is None:
        keys = list(FH_BYTE2SAMPLE.keys())
    start = f.tell()
    f.seek(0)
    text_header = f.read(3600)
    bfh = BinaryFileHeader()
    for k in keys:
        offset = FH_BYTE2SAMPLE[k]
        # all fields are 2 or 4 bytes integers
        size = 4 if k in ("Job", "Line", "Reel") else 2
        fmt = ">i" if size == 4 else ">h"
        if not bigendian:
            fmt = "<i" if size == 4 else "<h"
        val_bytes = text_header[offset:offset+size]
        val = struct.unpack(fmt, val_bytes)[0]
        setattr(bfh, k, val)
    bfh.keys_loaded = list(keys)
    f.seek(start)
    return FileHeader(text_header[:3200], bfh)


def read_traceheader(
    f: BinaryIO, keys: Optional[Iterable[str]] = None, bigendian: bool = True
) -> BinaryTraceHeader:
    """
    Read a single binary trace header from ``f``.

    Parameters
    ----------
    f : BinaryIO
        Open binary file handle positioned at a trace header.
    keys : Iterable[str], optional
        Header fields to read; all are loaded when omitted.
    bigendian : bool, optional
        ``True`` for big-endian encoding.

    Returns
    -------
    BinaryTraceHeader
        Parsed header object.
    """
    if keys is None:
        keys = list(TH_BYTE2SAMPLE.keys())
    hdr_bytes = f.read(240)
    th = BinaryTraceHeader()
    for k in keys:
        offset, size = TH_BYTE2SAMPLE[k]
        fmt = ">i" if size == 4 else ">h"
        if not bigendian:
            fmt = "<i" if size == 4 else "<h"
        val = struct.unpack(fmt, hdr_bytes[offset:offset + size])[0]
        setattr(th, k, val)
    th.keys_loaded = list(keys)
    return th


def read_traces(
    f: BinaryIO,
    ns: int,
    ntraces: int,
    datatype: int,
    keys: Optional[Iterable[str]] = None,
    bigendian: bool = True,
) -> Tuple[List[BinaryTraceHeader], List[List[float]]]:
    """
    Read ``ntraces`` traces and their headers from ``f``.

    Parameters
    ----------
    f : BinaryIO
        Open file handle positioned at the first trace.
    ns : int
        Number of samples per trace.
    ntraces : int
        Number of traces to read.
    datatype : int
        SEGY data sample format code.
    keys : Iterable[str], optional
        Header fields to read for each trace.
    bigendian : bool, optional
        ``True`` for big-endian encoding.

    Returns
    -------
    tuple
        ``(headers, data)`` where ``headers`` is a list of
        :class:`BinaryTraceHeader` and ``data`` is ``ns`` x ``ntraces`` array.
    """
    data: np.ndarray = np.zeros((ns, ntraces), dtype=np.float32)
    headers: List[BinaryTraceHeader] = [
        BinaryTraceHeader() for _ in range(ntraces)
    ]

    trace_size = 240 + ns * 4
    raw = f.read(trace_size * ntraces)

    if keys is None:
        keys = list(TH_BYTE2SAMPLE.keys())
    key_list = list(keys)

    # Pre-compute unpack formats for header fields
    endian_char = ">" if bigendian else "<"
    hdr_parsers = []
    for k in key_list:
        offset_k, size = TH_BYTE2SAMPLE[k]
        fmt = endian_char + ("i" if size == 4 else "h")
        hdr_parsers.append((k, offset_k, fmt))

    ibm = datatype == 1

    def parse_one(idx: int):
        offset = idx * trace_size
        hdr_buf = raw[offset:offset + 240]
        hdr = BinaryTraceHeader()
        for k, offset_k, fmt in hdr_parsers:
            val = struct.unpack_from(fmt, hdr_buf, offset_k)[0]
            setattr(hdr, k, val)
        hdr.keys_loaded = key_list

        data_buf = raw[offset + 240:offset + trace_size]
        if ibm:
            samples = ibm_to_ieee_array(data_buf, ns, bigendian)
        else:
            samples = np.frombuffer(data_buf, dtype=endian_char + "f4", count=ns)
        return idx, hdr, samples

    with ThreadPoolExecutor() as pool:
        for idx, hdr, samples in pool.map(parse_one, range(ntraces)):
            headers[idx] = hdr
            data[:, idx] = samples

    return headers, data


def read_file(
    f: BinaryIO,
    warn_user: bool = True,
    keys: Optional[Iterable[str]] = None,
    bigendian: bool = True,
    workers: int = 5,
) -> SeisBlock:
    """
    Read a complete SEGY file from an open file handle.

    Parameters
    ----------
    f : BinaryIO
        File object to read from.
    warn_user : bool, optional
        Currently unused.
    keys : Iterable[str], optional
        Additional header fields to load with each trace.
    bigendian : bool, optional
        Set ``True`` for big-endian encoding.
    workers : int, optional
        Unused parameter kept for backwards compatibility.

    Returns
    -------
    SeisBlock
        Entire dataset loaded into memory. The file is read in chunks of
        ``TRACE_CHUNKSIZE`` traces to limit peak memory usage.
    """
    fh = read_fileheader(f, bigendian=bigendian)
    ns = fh.bfh.ns
    dsf = fh.bfh.DataSampleFormat
    trace_size = 240 + ns * 4
    f.seek(0, 2)
    end = f.tell()
    ntraces = (end - 3600) // trace_size
    f.seek(3600)
    headers: List[BinaryTraceHeader] = [BinaryTraceHeader() for _ in range(ntraces)]
    data: np.ndarray = np.zeros((ns, ntraces), dtype=np.float32)

    idx = 0
    while idx < ntraces:
        count = min(TRACE_CHUNKSIZE, ntraces - idx)
        h, d = read_traces(
            f, ns, count, dsf, keys, bigendian
        )
        for j in range(count):
            headers[idx + j] = h[j]
            data[:, idx + j] = d[:, j]
        idx += count

    return SeisBlock(fh, headers, data)


def segy_read(
    path: str,
    keys: Optional[Iterable[str]] = None,
    workers: int = 5,
    fs=None,
) -> SeisBlock:
    """
    Convenience wrapper to read a SEGY file.

    Parameters
    ----------
    path : str
        File system path to the SEGY file. When ``fs`` is provided the
        path is interpreted relative to that filesystem.
    fs : filesystem-like object, optional
        Filesystem providing ``open`` if reading from non-local storage.
    keys : Iterable[str], optional
        Additional header fields to load with each trace.

    Returns
    -------
    SeisBlock
        Loaded dataset.
    """
    print(f"Reading SEGY file {path}")

    opener = fs.open if fs is not None else open

    with opener(path, "rb") as f:
        block = read_file(f, keys=keys, workers=workers)
    print(
        f"Loaded header ns={block.fileheader.bfh.ns} "
        f"dt={block.fileheader.bfh.dt} from {path}"
    )
    return block



import os
from io import BytesIO
import gzip
import urllib.request
import shutil
import fsspec
import numpy as np

import pysegy as seg  # noqa: E402
from pysegy.ibm import ibm_to_ieee  # noqa: E402
from pysegy.types import FileHeader, BinaryTraceHeader, SeisBlock  # noqa: E402

DATAFILE = os.path.join(
    os.path.dirname(__file__), "..", "..", "data",
    "overthrust_2D_shot_1_20.segy",
)


def test_read():
    block = seg.segy_read(DATAFILE)
    assert block.fileheader.bfh.ns == 751
    assert len(block.traceheaders) == 3300
    assert block.traceheaders[0].SourceX == 400
    assert block.traceheaders[0].GroupX == 100


def test_write_roundtrip(tmp_path):
    fh = FileHeader()
    fh.bfh.ns = 4
    fh.bfh.DataSampleFormat = 5
    headers = [BinaryTraceHeader() for _ in range(2)]
    for th in headers:
        th.ns = 4
        th.SourceX = 1234
    data = np.array([[float(i*j) for j in range(2)] for i in range(4)])
    block = SeisBlock(fh, headers, data)
    tmp = tmp_path / 'temp.segy'
    seg.segy_write(str(tmp), block)
    out = seg.segy_read(str(tmp))
    assert out.fileheader.bfh.ns == 4
    assert out.traceheaders[0].SourceX == 1234
    assert np.all(out.data == data)


def test_ibm_conversion():
    """Ensure IBM -> IEEE conversion works for known constant."""
    assert ibm_to_ieee(b"\x41\x10\x00\x00") == 1.0


def test_fileheader_io():
    """Round-trip a file header using in-memory bytes."""
    fh = FileHeader()
    fh.bfh.Job = 99
    fh.bfh.Line = 123
    buf = BytesIO()
    seg.write.write_fileheader(buf, fh)
    buf.seek(0)
    out = seg.read.read_fileheader(buf)
    assert out.bfh.Job == 99
    assert out.bfh.Line == 123
    assert len(buf.getvalue()) == 3600


def test_write_read_block_bytesio():
    """Write and read a simple block using BytesIO."""
    fh = FileHeader()
    fh.bfh.ns = 2
    fh.bfh.DataSampleFormat = 5
    headers = [BinaryTraceHeader() for _ in range(1)]
    headers[0].ns = 2
    data = np.array([[1.0], [2.0]])
    block = SeisBlock(fh, headers, data)
    bio = BytesIO()
    seg.write.write_block(bio, block)
    bio.seek(0)
    out = seg.read.read_file(bio)
    assert np.all(out.data == data)


def test_read_with_filesystem():
    fs = fsspec.filesystem("file")
    block = seg.segy_read(DATAFILE, fs=fs)
    assert block.fileheader.bfh.ns == 751


def test_write_with_filesystem(tmp_path):
    fs = fsspec.filesystem("file")
    fh = FileHeader()
    fh.bfh.ns = 2
    fh.bfh.DataSampleFormat = 5
    hdr = BinaryTraceHeader()
    hdr.ns = 2
    hdr.SourceX = 111
    data = np.array([[1.0], [2.0]])
    block = SeisBlock(fh, [hdr], data)
    dest = tmp_path / "fsout.segy"
    seg.segy_write(str(dest), block, fs=fs)
    out = seg.segy_read(str(dest), fs=fs)
    assert out.fileheader.bfh.ns == 2
    assert out.traceheaders[0].SourceX == 111


BP_URL = (
    "http://s3.amazonaws.com/open.source.geoscience/"
    "open_data/bpmodel94/Model94_shots.segy.gz"
)


def test_bp_model_headers():
    """Download a portion of the BP model data and verify header values."""
    response = urllib.request.urlopen(BP_URL)
    with gzip.GzipFile(fileobj=response) as gz:
        data = gz.read(40000)
    fh = seg.read.read_fileheader(BytesIO(data))
    assert fh.bfh.dt == 4000
    assert fh.bfh.ns == 2000
    assert fh.bfh.DataSampleFormat == 1
    bio = BytesIO(data)
    bio.seek(3600)
    th = seg.read.read_traceheader(bio)
    assert th.ns == 2000
    assert th.SourceX == 0
    assert th.GroupX == 15


def test_bp_model_scan(tmp_path):
    """Download the full BP Model dataset and verify shot statistics.

    The dataset contains 278 distinct shot locations. Receiver counts vary
    between 240 and 480 per shot. This test ensures the reader can process the
    entire file and that these counts match the known reference values.
    """
    dest = tmp_path / "Model94_shots.segy"
    response = urllib.request.urlopen(BP_URL)
    with gzip.GzipFile(fileobj=response) as gz, open(dest, "wb") as f:
        shutil.copyfileobj(gz, f)

    scan = seg.segy_scan(str(dest))

    fh = scan.fileheader
    shots = scan.shots
    counts = scan.counts

    ns = fh.bfh.ns
    trace_size = 240 + ns * 4
    with open(dest, "rb") as f:
        f.seek(0, os.SEEK_END)
        total = (f.tell() - 3600) // trace_size

    assert total == int(sum(counts))
    assert len(shots) == 278
    assert int(min(counts)) == 240
    assert int(max(counts)) == 480

    hdrs = scan.read_headers(0, keys=["GroupX"])
    assert hdrs[0].GroupX == 15
    assert len(hdrs) == counts[0]


def test_scan_directory_pattern():
    data_dir = os.path.join(
        os.path.dirname(__file__), "..", "..", "data"
    )
    scan = seg.segy_scan(
        data_dir, "overthrust_2D_shot_*.segy", keys=["GroupX"]
    )
    assert isinstance(scan, seg.SegyScan)
    assert len(scan.shots) == 97
    assert len(set(scan.paths)) == 5
    idx = 0  # first shot across all files
    assert scan.paths[idx].endswith("overthrust_2D_shot_1_20.segy")
    assert scan.counts[idx] == 127
    assert scan.summary(idx)["GroupX"] == (100, 6400)


def test_scan_with_filesystem():
    fs = fsspec.filesystem("file")
    data_dir = os.path.join(
        os.path.dirname(__file__), "..", "..", "data"
    )
    scan = seg.segy_scan(
        data_dir,
        "overthrust_2D_shot_*.segy",
        keys=["GroupX"],
        fs=fs,
    )
    assert isinstance(scan, seg.SegyScan)
    assert len(scan.shots) == 97


def test_scan_unsorted_traces(tmp_path):
    """Ensure scanning handles files with interleaved shots."""
    fh = FileHeader()
    fh.bfh.ns = 1
    fh.bfh.DataSampleFormat = 5

    hdr1 = BinaryTraceHeader()
    hdr1.ns = 1
    hdr1.SourceX = 1
    hdr1.SourceY = 1
    hdr1.GroupX = 1

    hdr2 = BinaryTraceHeader()
    hdr2.ns = 1
    hdr2.SourceX = 2
    hdr2.SourceY = 2
    hdr2.GroupX = 2

    hdr3 = BinaryTraceHeader()
    hdr3.ns = 1
    hdr3.SourceX = 1
    hdr3.SourceY = 1
    hdr3.GroupX = 3

    headers = [hdr1, hdr2, hdr3]
    data = np.zeros((1, 3), dtype=np.float32)
    block = SeisBlock(fh, headers, data)
    tmp = tmp_path / "unsorted.segy"
    with open(tmp, "wb") as f:
        seg.write.write_block(f, block)

    scan = seg.segy_scan(str(tmp), keys=["GroupX"])
    assert len(scan.shots) == 2
    assert scan.shots[0] == (1.0, 1.0, 0.0)
    assert scan.counts == [2, 1]
    assert scan.summary(0)["GroupX"] == (1, 3)


def test_save_and_load_scan(tmp_path):
    scan = seg.segy_scan(DATAFILE)
    dest = tmp_path / "scan.pkl"
    seg.save_scan(str(dest), scan)
    out = seg.load_scan(str(dest))
    assert isinstance(out, seg.SegyScan)
    assert out.shots == scan.shots
    assert out.counts == scan.counts


def test_save_and_load_scan_fs(tmp_path):
    fs = fsspec.filesystem("file")
    scan = seg.segy_scan(DATAFILE)
    dest = tmp_path / "scan_fs.pkl"
    seg.save_scan(str(dest), scan, fs=fs)
    out = seg.load_scan(str(dest), fs=fs)
    assert out.shots == scan.shots


def test_index_and_lazy_data():
    scan = seg.segy_scan(DATAFILE)
    rec = scan[0]
    assert rec.coordinates == scan.shots[0]
    assert rec._data is None
    rec.data
    assert rec.data is not None
    assert rec._data is not None
    assert rec.fileheader.bfh.ns == scan.fileheader.bfh.ns
    all_blocks = scan.data
    assert len(all_blocks) == len(scan.shots)


def test_rec_coordinates():
    scan = seg.segy_scan(DATAFILE)
    rec = scan[0]
    coords = rec.rec_coordinates
    assert coords.shape[0] == scan.counts[0]
    assert tuple(coords[0]) == (100.0, 0.0, 0.0)


def test_get_header_scaling():
    fh = FileHeader()
    fh.bfh.ns = 1
    fh.bfh.DataSampleFormat = 5

    h1 = BinaryTraceHeader()
    h1.ns = 1
    h1.SourceX = 10
    h1.RecSourceScalar = 2

    h2 = BinaryTraceHeader()
    h2.ns = 1
    h2.SourceX = 20
    h2.RecSourceScalar = -2

    h3 = BinaryTraceHeader()
    h3.ns = 1
    h3.SourceX = 5
    h3.RecSourceScalar = 1

    h4 = BinaryTraceHeader()
    h4.ns = 1
    h4.SourceX = 7
    h4.RecSourceScalar = 0

    headers = [h1, h2, h3, h4]
    block = SeisBlock(fh, headers, np.zeros((1, 4), dtype=np.float32))

    vals = seg.get_header(block, "SourceX")
    assert vals[:2] == [20, 10]
    assert vals[2:] == [5, 7]

    raw = seg.get_header(block, "SourceX", scale=False)
    assert raw == [10, 20, 5, 7]



"""
Conversion helpers between IBM and IEEE floating point formats.
"""

from typing import Union
import numpy as np


def ibm_to_ieee(value: Union[bytes, bytearray, int]) -> float:
    """
    Convert a 4-byte IBM floating point number to ``float``.

    Parameters
    ----------
    value : bytes or int
        Input value encoded in the IBM 32-bit floating point format.

    Returns
    -------
    float
        Floating point number in IEEE representation.
    """
    if isinstance(value, (bytes, bytearray)):
        if len(value) != 4:
            raise ValueError("IBM float must be 4 bytes")
        val = int.from_bytes(value, byteorder='big', signed=False)
    else:
        val = value & 0xffffffff
    if val == 0:
        return 0.0
    sign = 1 if (val >> 31) == 0 else -1
    exponent = (val >> 24) & 0x7f
    fraction = val & 0x00ffffff
    # IBM exponent is base 16 biased by 64
    mant = fraction / float(0x01000000)
    return sign * mant * 16 ** (exponent - 64)


def ibm_to_ieee_array(buf: bytes, count: int, bigendian: bool = True) -> np.ndarray:
    """Vectorized conversion of ``count`` IBM floats contained in ``buf``."""
    dtype = ">u4" if bigendian else "<u4"
    vals = np.frombuffer(buf, dtype=dtype, count=count)
    sign = np.where(vals >> 31 == 0, 1.0, -1.0)
    exponent = (vals >> 24) & 0x7F
    fraction = vals & 0x00FFFFFF
    mant = fraction.astype(np.float64) / float(0x01000000)
    out = sign * mant * np.power(16.0, exponent.astype(np.float64) - 64)
    out[vals == 0] = 0.0
    return out.astype(np.float32)


def ieee_to_ibm(f: float) -> bytes:
    """
    Convert ``float`` to IBM 32-bit floating point bytes.

    Parameters
    ----------
    f : float
        Numeric value in IEEE representation.

    Returns
    -------
    bytes
        The value encoded using the IBM format.
    """
    if f == 0.0:
        return b"\x00\x00\x00\x00"
    sign = 0
    if f < 0:
        sign = 0x80
        f = -f
    exponent = 64
    while f < 1.0:
        f *= 16.0
        exponent -= 1
    while f >= 16.0:
        f /= 16.0
        exponent += 1
    fraction = int(f * 0x01000000) & 0x00ffffff
    val = (sign << 24) | (exponent << 24) | fraction
    return val.to_bytes(4, byteorder='big')



"""Minimal Python port of SegyIO.jl."""
# flake8: noqa



from .types import (
    BinaryFileHeader,
    BinaryTraceHeader,
    FileHeader,
    SeisBlock,
)
from .read import (
    read_fileheader,
    read_traceheader,
    read_file,
    segy_read,
)
from .scan import (
    ShotRecord,
    SegyScan,
    segy_scan,
    save_scan,
    load_scan
)
from .write import (
    write_fileheader,
    write_traceheader,
    write_block,
    segy_write,
)
from .utils import get_header

__all__ = [
    "BinaryFileHeader",
    "BinaryTraceHeader",
    "FileHeader",
    "SeisBlock",
    "SegyScan",
    "ShotRecord",
    "read_fileheader",
    "read_traceheader",
    "read_file",
    "segy_read",
    "segy_scan",
    "save_scan",
    "load_scan",
    "write_fileheader",
    "write_traceheader",
    "write_block",
    "segy_write",
    "get_header",
]




from typing import Iterable, List, Union

from .types import BinaryTraceHeader, SeisBlock

_RECSRC_FIELDS = {
    "SourceX",
    "SourceY",
    "GroupX",
    "GroupY",
    "CDPX",
    "CDPY",
}

_ELEV_FIELDS = {
    "RecGroupElevation",
    "SourceSurfaceElevation",
    "SourceDepth",
    "RecDatumElevation",
    "SourceDatumElevation",
    "SourceWaterDepth",
    "GroupWaterDepth",
}


def _check_scale(name: str) -> tuple[bool, str]:
    if name in _RECSRC_FIELDS:
        return True, "RecSourceScalar"
    if name in _ELEV_FIELDS:
        return True, "ElevationScalar"
    return False, ""


def get_header(
    src: Union[SeisBlock, Iterable[BinaryTraceHeader]],
    name: str,
    *,
    scale: bool = True,
) -> List[float]:
    """Return values for ``name`` from ``src`` optionally applying scaling."""
    if isinstance(src, SeisBlock):
        headers = src.traceheaders
    else:
        headers = list(src)

    vals = [getattr(h, name) for h in headers]

    scalable, scale_name = _check_scale(name)
    if scale and scalable:
        scaled: List[float] = []
        for h, v in zip(headers, vals):
            fact = getattr(h, scale_name)
            if fact > 0:
                scaled.append(v * fact)
            elif fact < 0:
                scaled.append(v / abs(fact))
            else:
                scaled.append(v)
        return scaled
    return vals


__all__ = ["get_header"]



import os

import pysegy as seg  # noqa: E402

DATAFILE = os.path.join(
    os.path.dirname(__file__),
    "..",
    "..",
    "data",
    "overthrust_2D_shot_1_20.segy",
)


def test_segy_scan_directory_pattern():
    data_dir = os.path.join(os.path.dirname(__file__), "..", "..", "data")
    scan = seg.segy_scan(
        data_dir,
        "overthrust_2D_shot_*.segy",
        keys=["GroupX"],
    )
    assert isinstance(scan, seg.SegyScan)
    assert len(scan.shots) == 97
    assert len(set(scan.paths)) == 5
    idx = 0
    assert scan.paths[idx].endswith("overthrust_2D_shot_1_20.segy")
    assert scan.counts[idx] == 127
    assert scan.summary(idx)["GroupX"] == (100, 6400)



"""
Writing utilities for the minimal Python SEGY implementation.
"""

import struct
from typing import BinaryIO
from .types import (
    SeisBlock,
    FileHeader,
    BinaryTraceHeader,
    FH_BYTE2SAMPLE,
    TH_BYTE2SAMPLE,
)
from .ibm import ieee_to_ibm


def write_fileheader(
    f: BinaryIO, fh: FileHeader, bigendian: bool = True
) -> None:
    """
    Write ``fh`` to ``f``.

    Parameters
    ----------
    f : BinaryIO
        Open binary file handle to write to.
    fh : FileHeader
        Header values to write.
    bigendian : bool, optional
        Write numbers in big-endian order when ``True``.
    """
    th = fh.th
    if len(th) < 3200:
        th = th + b" " * (3200 - len(th))
    f.write(th[:3200])
    bfh = fh.bfh
    size_written = 3200
    for key in FH_BYTE2SAMPLE:
        val = getattr(bfh, key)
        size = 4 if key in ("Job", "Line", "Reel") else 2
        fmt = ">i" if size == 4 else ">h"
        if not bigendian:
            fmt = "<i" if size == 4 else "<h"
        f.write(struct.pack(fmt, val))
        size_written += size
    pad = 3600 - size_written
    if pad > 0:
        f.write(b"\x00" * pad)


def write_traceheader(
    f: BinaryIO, th: BinaryTraceHeader, bigendian: bool = True
) -> None:
    """
    Write a single trace header to ``f``.

    Parameters
    ----------
    f : BinaryIO
        Open file handle for output.
    th : BinaryTraceHeader
        Trace header values to write.
    bigendian : bool, optional
        Write numbers in big-endian order when ``True``.
    """
    buf = bytearray(240)
    for key in TH_BYTE2SAMPLE:
        val = getattr(th, key)
        offset, size = TH_BYTE2SAMPLE[key]
        fmt = ">i" if size == 4 else ">h"
        if not bigendian:
            fmt = "<i" if size == 4 else "<h"
        struct.pack_into(fmt, buf, offset, val)
    f.write(buf)


def write_block(f: BinaryIO, block: SeisBlock, bigendian: bool = True) -> None:
    """
    Write an entire :class:`SeisBlock` to ``f``.

    Parameters
    ----------
    f : BinaryIO
        File handle where the block will be written.
    block : SeisBlock
        Data to serialise.
    bigendian : bool, optional
        Write numbers in big-endian order when ``True``.
    """
    write_fileheader(f, block.fileheader, bigendian)
    ns = block.fileheader.bfh.ns
    dsf = block.fileheader.bfh.DataSampleFormat
    for i, hdr in enumerate(block.traceheaders):
        trace = block.data[:, i]
        write_traceheader(f, hdr, bigendian)
        if dsf == 1:
            converted = b"".join(ieee_to_ibm(float(x)) for x in trace)
            f.write(converted)
        else:
            fmt = ">%df" % ns if bigendian else "<%df" % ns
            f.write(struct.pack(fmt, *trace))


def segy_write(path: str, block: SeisBlock, fs=None) -> None:
    """
    Convenience wrapper to write ``block`` to ``path``.

    Parameters
    ----------
    path : str
        Destination file path. When ``fs`` is provided the path is
        interpreted relative to that filesystem.
    block : SeisBlock
        Dataset to write to disk.
    """
    print(f"Writing SEGY file {path}")

    opener = fs.open if fs is not None else open

    with opener(path, "wb") as f:
        write_block(f, block)
    print(f"Finished writing {path}")



"""
Shared data structures for the minimal Python implementation.
"""

from dataclasses import dataclass, field
from typing import Dict, List, Tuple

# Byte locations for binary file header fields
FH_BYTE2SAMPLE: Dict[str, int] = {
    "Job": 3200,
    "Line": 3204,
    "Reel": 3208,
    "DataTracePerEnsemble": 3212,
    "AuxiliaryTracePerEnsemble": 3214,
    "dt": 3216,
    "dtOrig": 3218,
    "ns": 3220,
    "nsOrig": 3222,
    "DataSampleFormat": 3224,
    "EnsembleFold": 3226,
    "TraceSorting": 3228,
    "VerticalSumCode": 3230,
    "SweepFrequencyStart": 3232,
    "SweepFrequencyEnd": 3234,
    "SweepLength": 3236,
    "SweepType": 3238,
    "SweepChannel": 3240,
    "SweepTaperlengthStart": 3242,
    "SweepTaperLengthEnd": 3244,
    "TaperType": 3246,
    "CorrelatedDataTraces": 3248,
    "BinaryGain": 3250,
    "AmplitudeRecoveryMethod": 3252,
    "MeasurementSystem": 3254,
    "ImpulseSignalPolarity": 3256,
    "VibratoryPolarityCode": 3258,
    "SegyFormatRevisionNumber": 3500,
    "FixedLengthTraceFlag": 3502,
    "NumberOfExtTextualHeaders": 3504,
}

# Byte ranges for trace header fields. Each entry maps a header name to a
# ``(offset, size)`` tuple where ``size`` is either 2 or 4 bytes depending on
# the SEGY specification.
_TH_OFFSETS: Dict[str, int] = {
    "TraceNumWithinLine": 0,
    "TraceNumWithinFile": 4,
    "FieldRecord": 8,
    "TraceNumber": 12,
    "EnergySourcePoint": 16,
    "CDP": 20,
    "CDPTrace": 24,
    "TraceIDCode": 28,
    "NSummedTraces": 30,
    "NStackedTraces": 32,
    "DataUse": 34,
    "Offset": 36,
    "RecGroupElevation": 40,
    "SourceSurfaceElevation": 44,
    "SourceDepth": 48,
    "RecDatumElevation": 52,
    "SourceDatumElevation": 56,
    "SourceWaterDepth": 60,
    "GroupWaterDepth": 64,
    "ElevationScalar": 68,
    "RecSourceScalar": 70,
    "SourceX": 72,
    "SourceY": 76,
    "GroupX": 80,
    "GroupY": 84,
    "CoordUnits": 88,
    "WeatheringVelocity": 90,
    "SubWeatheringVelocity": 92,
    "UpholeTimeSource": 94,
    "UpholeTimeGroup": 96,
    "StaticCorrectionSource": 98,
    "StaticCorrectionGroup": 100,
    "TotalStaticApplied": 102,
    "LagTimeA": 104,
    "LagTimeB": 106,
    "DelayRecordingTime": 108,
    "MuteTimeStart": 110,
    "MuteTimeEnd": 112,
    "ns": 114,
    "dt": 116,
    "GainType": 118,
    "InstrumentGainConstant": 120,
    "InstrumntInitialGain": 122,
    "Correlated": 124,
    "SweepFrequencyStart": 126,
    "SweepFrequencyEnd": 128,
    "SweepLength": 130,
    "SweepType": 132,
    "SweepTraceTaperLengthStart": 134,
    "SweepTraceTaperLengthEnd": 136,
    "TaperType": 138,
    "AliasFilterFrequency": 140,
    "AliasFilterSlope": 142,
    "NotchFilterFrequency": 144,
    "NotchFilterSlope": 146,
    "LowCutFrequency": 148,
    "HighCutFrequency": 150,
    "LowCutSlope": 152,
    "HighCutSlope": 154,
    "Year": 156,
    "DayOfYear": 158,
    "HourOfDay": 160,
    "MinuteOfHour": 162,
    "SecondOfMinute": 164,
    "TimeCode": 166,
    "TraceWeightingFactor": 168,
    "GeophoneGroupNumberRoll": 170,
    "GeophoneGroupNumberTraceStart": 172,
    "GeophoneGroupNumberTraceEnd": 174,
    "GapSize": 176,
    "OverTravel": 178,
    "CDPX": 180,
    "CDPY": 184,
    "Inline3D": 188,
    "Crossline3D": 192,
    "ShotPoint": 196,
    "ShotPointScalar": 200,
    "TraceValueMeasurmentUnit": 202,
    "TransductionConstnatMantissa": 204,
    "TransductionConstantPower": 208,
    "TransductionUnit": 210,
    "TraceIdentifier": 212,
    "ScalarTraceHeader": 214,
    "SourceType": 216,
    "SourceEnergyDirectionMantissa": 218,
    "SourceEnergyDirectionExponent": 222,
    "SourceMeasurmentMantissa": 224,
    "SourceMeasurementExponent": 228,
    "SourceMeasurmentUnit": 230,
    "Unassigned1": 232,
    "Unassigned2": 236,
}

# Header fields stored as 4-byte integers
_TH_INT32_FIELDS = {
    "TraceNumWithinLine",
    "TraceNumWithinFile",
    "FieldRecord",
    "TraceNumber",
    "EnergySourcePoint",
    "CDP",
    "CDPTrace",
    "Offset",
    "RecGroupElevation",
    "SourceSurfaceElevation",
    "SourceDepth",
    "RecDatumElevation",
    "SourceDatumElevation",
    "SourceWaterDepth",
    "GroupWaterDepth",
    "SourceX",
    "SourceY",
    "GroupX",
    "GroupY",
    "CDPX",
    "CDPY",
    "Inline3D",
    "Crossline3D",
    "ShotPoint",
    "TransductionConstnatMantissa",
    "SourceEnergyDirectionMantissa",
    "SourceMeasurmentMantissa",
    "Unassigned1",
    "Unassigned2",
}

# Final mapping of header field to (offset, size)
TH_BYTE2SAMPLE: Dict[str, Tuple[int, int]] = {
    k: (off, 4 if k in _TH_INT32_FIELDS else 2)
    for k, off in _TH_OFFSETS.items()
}

# Discard private constants from namespace
del _TH_OFFSETS
del _TH_INT32_FIELDS

FH_FIELDS = list(FH_BYTE2SAMPLE.keys())
TH_FIELDS = list(TH_BYTE2SAMPLE.keys())


@dataclass
class BinaryFileHeader:
    """
    Container for parsed binary file header values.
    """

    values: Dict[str, int] = field(
        default_factory=lambda: {k: 0 for k in FH_FIELDS}
    )
    keys_loaded: List[str] = field(default_factory=lambda: list(FH_FIELDS))

    def __getattr__(self, name):
        """Return the value for ``name`` from ``values``."""
        if name in self.values:
            return self.values[name]
        raise AttributeError(name)

    def __setattr__(self, name, value):
        """Set ``name`` in ``values`` when it is a header field."""
        if name in {"values", "keys_loaded"} or name not in FH_FIELDS:
            super().__setattr__(name, value)
        else:
            self.values[name] = int(value)
            if name not in self.keys_loaded:
                self.keys_loaded.append(name)

    def __str__(self):
        """Return a multi-line representation of loaded fields."""
        lines = ["BinaryFileHeader:"]
        fields = self.keys_loaded or FH_FIELDS
        for k in fields:
            lines.append(f"    {k:30s}: {self.values[k]:9d}")
        return "\n".join(lines)

    __repr__ = __str__

    def __getstate__(self):
        return {"values": self.values, "keys_loaded": self.keys_loaded}

    def __setstate__(self, state):
        super().__setattr__("values", state["values"])
        super().__setattr__("keys_loaded", state["keys_loaded"])


@dataclass
class FileHeader:
    """
    Combined textual and binary file header.
    """

    th: bytes = b" " * 3200
    bfh: BinaryFileHeader = field(default_factory=BinaryFileHeader)

    def __str__(self) -> str:
        """Return a readable representation of the header."""
        return str(self.bfh)

    __repr__ = __str__


@dataclass
class BinaryTraceHeader:
    """
    Container for parsed binary trace header values.
    """

    values: Dict[str, int] = field(
        default_factory=lambda: {k: 0 for k in TH_FIELDS}
    )
    keys_loaded: List[str] = field(default_factory=list)

    def __getattr__(self, name):
        """Return the header value ``name``."""
        if name in self.values:
            return self.values[name]
        raise AttributeError(name)

    def __setattr__(self, name, value):
        """Assign ``value`` to ``name`` when valid."""
        if name in {"values", "keys_loaded"} or name not in TH_FIELDS:
            super().__setattr__(name, value)
        else:
            self.values[name] = int(value)
            if name not in self.keys_loaded:
                self.keys_loaded.append(name)

    def __repr__(self):
        return self.__str__()

    def __getstate__(self):
        return {"values": self.values, "keys_loaded": self.keys_loaded}

    def __setstate__(self, state):
        super().__setattr__("values", state["values"])
        super().__setattr__("keys_loaded", state["keys_loaded"])

    def __str__(self):
        """Return a multi-line representation of loaded fields."""
        lines = ["BinaryTraceHeader:"]
        fields = self.keys_loaded or TH_FIELDS
        for k in fields:
            lines.append(f"    {k:30s}: {self.values[k]:9d}")
        return "\n".join(lines)


@dataclass
class SeisBlock:
    """
    In-memory representation of a SEGY dataset.
    """

    fileheader: FileHeader
    traceheaders: List[BinaryTraceHeader]
    data: List[List[float]]

    def __len__(self) -> int:
        return len(self.traceheaders)

    def __str__(self) -> str:
        lines = ["SeisBlock:"]
        lines.append(f"    traces: {len(self.traceheaders)}")
        lines.append(f"    ns: {self.fileheader.bfh.ns}")
        lines.append(f"    dt: {self.fileheader.bfh.dt}")
        return "\n".join(lines)

    __repr__ = __str__
